#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Definitions of different operations on a DataFrame."""

from typing import TYPE_CHECKING, List

from tqdm import tqdm

from graviti.openapi import add_data, update_data, update_schema, upload_files
from graviti.operation.common import get_schema
from graviti.portex import PortexType
from graviti.utility import File, chunked

if TYPE_CHECKING:
    from graviti.dataframe import DataFrame, FileSeries

_MAX_BATCH_SIZE = 2048
_MAX_ITEMS = 60000


class DataFrameOperation:
    """This class defines the basic method of the operation on a DataFrame."""

    def get_file_arrays(self) -> List["FileSeries"]:  # pylint: disable=no-self-use
        """Get the list of FileSeries.

        Returns:
            The list of FileSeries.

        """
        return []

    def get_upload_count(self) -> int:  # pylint: disable=no-self-use
        """Get the data amount to be uploaded.

        Returns:
            The data amount to be uploaded.

        """
        return 0

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI create sheet.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class DataOperation(DataFrameOperation):  # pylint: disable=abstract-method
    """This class defines the basic method of the data operation on a DataFrame."""

    _file_arrays: List["FileSeries"]

    def __init__(self, data: "DataFrame") -> None:
        self._data = data

    def _get_max_batch_size(self) -> int:
        return min(
            _MAX_BATCH_SIZE,
            _MAX_ITEMS // self._data.schema._get_column_count(),  # pylint: disable=protected-access
        )

    def get_file_arrays(self) -> List["FileSeries"]:
        """Get the list of FileSeries.

        Returns:
            The list of FileSeries.

        """
        if not hasattr(self, "_file_arrays"):
            self._file_arrays = self._data._get_file_columns()  # pylint: disable=protected-access

        return self._file_arrays

    def get_upload_count(self) -> int:
        """Get the data amount to be uploaded.

        Returns:
            The data amount to be uploaded.

        """
        return len(self._data)


class AddData(DataOperation):
    """This class defines the operation that add data to a DataFrame.

    Arguments:
        data: The data to be added.

    """

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        data = self._data._to_post_data()  # pylint: disable=protected-access

        for batch, *file_arrays in zip(
            *map(
                lambda x: chunked(x, self._get_max_batch_size()),  # type: ignore[arg-type]
                (data, *self.get_file_arrays()),
            )
        ):
            for file_array in file_arrays:
                local_files = filter(lambda x: isinstance(x, File), file_array)
                upload_files(
                    access_key,
                    url,
                    owner,
                    dataset,
                    draft_number=draft_number,
                    sheet=sheet,
                    files=local_files,
                    jobs=jobs,
                    pbar=file_pbar,
                )

            add_data(
                access_key,
                url,
                owner,
                dataset,
                draft_number=draft_number,
                sheet=sheet,
                data=batch,
            )
            data_pbar.update(len(batch))


class UpdateSchema(DataFrameOperation):
    """This class defines the operation that update the schema of a DataFrame.

    Arguments:
        schema: New portex schema after updated.

    """

    def __init__(self, schema: PortexType) -> None:
        self.schema = schema

    def do(  # pylint: disable=invalid-name, unused-argument
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI update schema.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        portex_schema, avro_schema, arrow_schema = get_schema(self.schema)

        update_schema(
            access_key,
            url,
            owner,
            dataset,
            draft_number=draft_number,
            sheet=sheet,
            _schema=portex_schema,
            _avro_schema=avro_schema,
            _arrow_schema=arrow_schema,
        )


class UpdateData(DataOperation):
    """This class defines the operation that updates the data of a DataFrame.

    Arguments:
        data: The data for updating.

    """

    _file_arrays: List["FileSeries"]

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
        data_pbar: tqdm,
        file_pbar: tqdm,
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.
            data_pbar: The process bar for uploading structured data.
            file_pbar: The process bar for uploading binary files.

        """
        data = self._data._to_patch_data()  # pylint: disable=protected-access

        for batch, *file_arrays in zip(
            *map(
                lambda x: chunked(x, self._get_max_batch_size()),  # type: ignore[arg-type]
                (data, *self.get_file_arrays()),
            )
        ):
            for file_array in file_arrays:
                local_files = filter(lambda x: isinstance(x, File), file_array)
                upload_files(
                    access_key,
                    url,
                    owner,
                    dataset,
                    draft_number=draft_number,
                    sheet=sheet,
                    files=local_files,
                    jobs=jobs,
                    pbar=file_pbar,
                )

            update_data(
                access_key,
                url,
                owner,
                dataset,
                draft_number=draft_number,
                sheet=sheet,
                data=batch,
            )
            data_pbar.update(len(batch))



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti Series."""


from itertools import islice
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.column.indexing import ColumnSeriesILocIndexer, ColumnSeriesLocIndexer
from graviti.dataframe.container import Container
from graviti.operation import UpdateData
from graviti.paging import LazyFactoryBase, PagingList, PyArrowPagingList
from graviti.paging.lists import PagingListBase
from graviti.utility import MAX_REPR_ROWS, FileBase

if TYPE_CHECKING:
    from graviti.dataframe.frame import DataFrame

_SB = TypeVar("_SB", bound="SeriesBase")
_S = TypeVar("_S", bound="Series")
_A = TypeVar("_A", bound="ArraySeries")
_E = TypeVar("_E", bound="EnumSeries")


class SeriesBase(Container):  # pylint: disable=abstract-method
    """One-dimensional array.

    Arguments:
        data: The data that needs to be stored in Series. Could be ndarray or Iterable.
        schema: Data type to force. If None, will be inferred from ``data``.

    Examples:
        Constructing Series from a list.

        >>> d = [1,2,3,4]
        >>> series = Series(data=d)
        >>> series
        0 1
        1 2
        2 3
        3 4

    """

    schema: pt.PortexType
    _data: PagingListBase

    def __new__(
        cls: Type[_SB],
        data: Union[Iterable[Any], _SB],
        schema: Optional[pt.PortexType] = None,
    ) -> Any:
        """One-dimensional data with schema.

        Arguments:
            data: The data that needs to be stored in column series.
            schema: The schema of the column series. If None, will be inferred from `data`.

        Raises:
            ValueError: When both given schema and Series data.
            ValueError: When using DataFrame as data to construct a Series.

        Returns:
            The created Series object.

        """
        if isinstance(data, cls):
            if schema is not None:
                raise ValueError("Only support the schema when data is not Series")

            return data.copy()

        if data.__class__.__name__ == "DataFrame":
            raise ValueError("Do not support Constructing a Series through a DataFrame")

        if schema is None:
            array = pa.array(data)
            schema = pt.PortexType.from_pyarrow(array.type)
        else:
            array = cls._pylist_to_pyarrow(data, schema)  # type: ignore[arg-type]

        return schema.container._from_pyarrow(array, schema)

    # @overload
    # def __getitem__(self, key: slice) -> "Series":
    #    ...

    # @overload
    # def __getitem__(self, key: int) -> Any:
    #     ...

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "Series":
    #     ...

    def __getitem__(self, key: int) -> Any:
        return self._data[key]

    @overload
    def __setitem__(self, key: slice, value: Union[Iterable[Any], _SB]) -> None:
        ...

    @overload
    def __setitem__(self, key: int, value: Any) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[Iterable[Any], _SB, Any],
    ) -> None:
        if isinstance(key, int):
            value = self._from_pyarrow(self._pylist_to_pyarrow([value], self.schema), self.schema)
            key = slice(key, key + 1)
        elif not isinstance(value, self.__class__):
            value = self._from_pyarrow(
                self._pylist_to_pyarrow(value, self.schema), self.schema  # type: ignore[arg-type]
            )
        elif not self.schema.to_pyarrow().equals(value.schema.to_pyarrow()):
            raise TypeError("The schema of the given Series is mismatched")

        self._set_item_by_slice(key, value)
        if self._root is not None and self._root.operations is not None:
            value = value.copy()
            root = self._root
            _record_key = root._record_key
            # TODO: support slicing methods for record_key
            record_key = [
                _record_key[i]  # type: ignore[index]
                for i in range(*key.indices(len(_record_key)))  # type: ignore[arg-type]
            ]
            name = self._name
            dataframe = root._construct(
                {name[0]: value},
                pt.record({name[0]: value.schema}),
                Series(record_key),
                name[1:],
            )
            root.operations.append(UpdateData(dataframe))  # type: ignore[union-attr]

    def __len__(self) -> int:
        return self._data.__len__()

    def __repr__(self) -> str:
        indices = list(self._get_repr_indices())
        indice_width = len(str(max(indices, default=1)))

        body = []
        body_item_width = 0
        for i in indices:
            item = self.loc[i]
            name = item._repr_folding() if hasattr(item, "_repr_folding") else str(item)
            body.append(name)
            body_item_width = max(len(name), body_item_width)

        lines = []
        for indice, value in zip(indices, body):
            lines.append(f"{indice:<{indice_width+2}}{value:<{body_item_width+2}}")
        if self.__len__() > MAX_REPR_ROWS:
            lines.append(f"...({self.__len__()})")
        return "\n".join(lines)

    def _repr_folding(self) -> str:
        return f"{self.__class__.__name__}({self.__len__()})"

    def _get_repr_indices(self) -> Iterable[int]:
        return islice(range(len(self)), MAX_REPR_ROWS)

    # @overload
    # @staticmethod
    # def _get_location_by_index(key: Iterable[int]) -> List[int]:
    #     ...

    # @overload
    # @staticmethod
    # def _get_location_by_index(key: int) -> int:
    #     ...

    # @staticmethod
    # def _get_location_by_index(key: Union[int, Iterable[int]]) -> Union[int, List[int]]:
    #     if isinstance(key, int):
    #         return key

    #     return list(key)

    # @overload
    # def _getitem_by_location(self, key: slice) -> "Series":
    #    ...

    # @overload
    # def _getitem_by_location(self, key: int) -> Any:
    #     ...

    # @overload
    # def _getitem_by_location(self, key: Iterable[int]) -> "Series":
    #     ...

    def _getitem_by_location(self, key: int) -> Any:
        return self.__getitem__(key)

    @staticmethod
    def _pylist_to_pyarrow(values: Iterable[Any], schema: pt.PortexType) -> pa.StructArray:

        processor, need_process = SeriesBase._process_array(values, schema)

        if not need_process:
            return pa.array(values, schema.to_pyarrow())

        return pa.array(processor(values), schema.to_pyarrow())

    @staticmethod
    def _process_array(
        values: Iterable[Any], schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        if not values:
            return lambda x: x, None

        for value in values:
            if value is None:
                continue

            processor, need_process = SeriesBase._get_process(value, schema)
            if need_process is None:
                continue

            return lambda x: list(map(processor, x)), need_process

        return lambda x: x, False

    @staticmethod
    def _get_process(
        value: Any, schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], Optional[bool]]:
        container = schema.container
        if value.__class__.__name__ == "DataFrame":
            return lambda x: x._to_post_data(), True  # pylint: disable=protected-access
        if container == ArraySeries:
            return SeriesBase._process_array(
                value, schema.to_builtin().items  # type: ignore[attr-defined]
            )
        if container == FileSeries:
            return SeriesBase._process_file(value)
        if container == EnumSeries:
            values_to_indices: Dict[Union[int, float, str, bool, None], Optional[int]] = {
                k: v for v, k in enumerate(schema.to_builtin().values)  # type: ignore[attr-defined]
            }
            if None not in values_to_indices:
                values_to_indices[None] = None
            return lambda x: values_to_indices[x], True

        return lambda x: x, False

    @staticmethod
    def _process_file(values: Iterable[Any]) -> Tuple[Callable[[Any], Any], bool]:
        if isinstance(values, FileBase):
            return lambda x: x.to_pyobj(), True

        return lambda x: x, False

    def _extend(self: _SB, values: _SB) -> None:
        """Extend Series to itself row by row.

        Arguments:
            values: A series that needs to be extended.

        Examples:
            >>> s1 = Series.from_pyarrow(pa.array([1,2,3]))
            >>> s1
            0  1
            1  2
            2  3
            >>> s2 = Series.from_pyarrow(pa.array([4,5,6]))
            >>> s2
            0  4
            1  5
            2  6
            >>> s1.extend(s2)
            >>> s1
            0  1
            1  2
            2  3
            3  1
            4  2
            5  3

        """
        self._data.extend(values._data)  # pylint: disable=protected-access

    def _copy(
        self: _SB,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _SB:
        obj: _SB = object.__new__(self.__class__)

        obj.schema = schema
        # pylint: disable=protected-access
        obj._root = root
        obj._data = self._data.copy()
        obj._name = name

        return obj

    def _set_item_by_slice(self, key: slice, value: _SB) -> None:
        self._data.set_slice(key, value._data)  # pylint: disable=protected-access

    @property
    def iloc(self) -> ColumnSeriesILocIndexer:
        """Purely integer-location based indexing for selection by position.

        Allowed inputs are:

        - An integer, e.g. ``5``.
        - A list or array of integers, e.g. ``[4, 3, 0]``.
        - A slice object with ints, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the ILocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return ColumnSeriesILocIndexer(self)

    @property
    def loc(self) -> ColumnSeriesLocIndexer:
        """Access a group of rows and columns by indexes or a boolean array.

        Allowed inputs are:

        - A single index, e.g. ``5``.
        - A list or array of indexes, e.g. ``[4, 3, 0]``.
        - A slice object with indexes, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the LocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return ColumnSeriesLocIndexer(self)

    @classmethod
    def from_pyarrow(
        cls: Type[_SB],
        array: pa.Array,
        schema: Optional[pt.PortexType] = None,
    ) -> _SB:
        """Instantiate a Series backed by an pyarrow array.

        Arguments:
            array: The input pyarrow array.
            schema: The schema of the series. If None, will be inferred from `array`.

        Raises:
            TypeError: When the schema is mismatched with the pyarrow array type.

        Returns:
            The loaded :class:`~graviti.dataframe.column.Series` instance.

        """
        if schema is None:
            schema = pt.PortexType.from_pyarrow(array.type)
        elif not array.type.equals(schema.to_pyarrow()):
            raise TypeError("The schema is mismatched with the pyarrow array")

        return cls._from_pyarrow(array, schema)

    def _to_post_data(self) -> List[Any]:
        return self.to_pylist()


@pt.ContainerRegister(
    pt.binary,
    pt.boolean,
    pt.float32,
    pt.float64,
    pt.int32,
    pt.int64,
    pt.string,
)
class Series(SeriesBase):  # pylint: disable=abstract-method
    """One-dimensional array.

    Arguments:
        data: The data that needs to be stored in Series. Could be ndarray or Iterable.
        schema: Data type to force. If None, will be inferred from ``data``.

    Examples:
        Constructing Series from a list.

        >>> d = [1,2,3,4]
        >>> series = Series(data=d)
        >>> series
        0 1
        1 2
        2 3
        3 4

    """

    _data: PyArrowPagingList

    def __getitem__(self, key: int) -> Any:
        return self._data[key].as_py()

    @classmethod
    def _from_factory(
        cls: Type[_S],
        factory: LazyFactoryBase,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _S:
        obj: _S = object.__new__(cls)
        obj._data = factory.create_pyarrow_list()
        obj.schema = schema
        obj._root = root
        obj._name = name
        return obj

    @classmethod
    def _from_pyarrow(
        cls: Type[_S],
        array: pa.Array,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _S:
        obj: _S = object.__new__(cls)
        obj._data = PyArrowPagingList.from_pyarrow(array)
        obj.schema = schema
        obj._root = root
        obj._name = name
        return obj

    def to_pylist(self) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        return self._data.to_pyarrow().to_pylist()  # type: ignore[no-any-return]


@pt.ContainerRegister(pt.array)
class ArraySeries(SeriesBase):  # pylint: disable=abstract-method
    """One-dimensional array for portex builtin type array."""

    _data: PagingList
    _item_schema: pt.PortexType

    def __getitem__(self, key: int) -> Any:
        return self._data[key]

    @classmethod
    def _from_factory(
        cls: Type[_A],
        factory: LazyFactoryBase,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _A:
        builtin_schema: pt.array = schema.to_builtin()  # type: ignore[assignment]
        _item_schema = builtin_schema.items
        _item_creator = _item_schema.container._from_pyarrow  # pylint: disable=protected-access

        obj: _A = object.__new__(cls)
        obj._data = factory.create_list(lambda scalar: _item_creator(scalar.values, _item_schema))
        obj.schema = schema
        obj._root = root
        obj._item_schema = _item_schema  # pylint: disable=protected-access
        obj._name = name

        return obj

    @classmethod
    def _from_pyarrow(
        cls: Type[_A],
        array: pa.Array,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _A:
        builtin_schema: pt.array = schema.to_builtin()  # type: ignore[assignment]
        _item_schema = builtin_schema.items
        _item_creator = _item_schema.container._from_pyarrow  # pylint: disable=protected-access

        obj: _A = object.__new__(cls)
        obj._data = PagingList(_item_creator(item.values, _item_schema) for item in array)

        obj.schema = schema
        obj._root = root
        obj._item_schema = _item_schema  # pylint: disable=protected-access
        obj._name = name

        return obj

    # TODO: copy method will be implementated sooner.
    # def copy(self: _A) -> _A:
    #     """Get a copy of the series.

    #     Returns:
    #         A copy of the series.

    #     """
    #     obj = super().copy()

    #     return obj

    def to_pylist(self) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        return [item.to_pylist() for item in self._data]

    def _to_post_data(self) -> List[Any]:
        return [item._to_post_data() for item in self._data]  # pylint: disable=protected-access


@pt.ExternalContainerRegister(
    "https://github.com/Project-OpenBytes/portex-standard",
    "main",
    "file.Audio",
    "file.Image",
    "file.PointCloud",
    "file.PointCloudBin",
    "file.RemoteFile",
    "label.file.RemoteInstanceMask",
    "label.file.RemoteSemanticMask",
)
class FileSeries(Series):  # pylint: disable=abstract-method
    """One-dimensional array for file."""

    def __getitem__(self, key: int) -> Optional[FileBase]:
        scalar = self._data[key]
        if not scalar.is_valid:
            return None

        return FileBase.from_pyarrow(scalar)


@pt.ContainerRegister(pt.enum)
class EnumSeries(Series):
    """One-dimensional array for portex builtin type enum."""

    _indices_to_values: Dict[Optional[pa.Scalar], Union[int, float, str, bool, None]]

    def __getitem__(self, key: int) -> Any:
        return self._indices_to_values[self._data[key]]

    def _to_post_data(self) -> List[int]:
        return self._data.to_pyarrow().to_pylist()  # type: ignore[no-any-return]

    def _copy(
        self,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> "EnumSeries":
        obj = super()._copy(schema, root, name)
        obj._indices_to_values = self._indices_to_values.copy()  # pylint: disable=protected-access
        return obj

    @classmethod
    def _from_factory(
        cls: Type[_E],
        factory: LazyFactoryBase,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _E:
        obj: _E = object.__new__(cls)
        obj._data = factory.create_pyarrow_list()
        obj.schema = schema
        obj._root = root
        obj._name = name
        obj._indices_to_values = cls._create_indices_to_values(
            schema.to_builtin().values,  # type: ignore[attr-defined]
            obj._data._patype,  # pylint: disable=protected-access
        )
        return obj

    @classmethod
    def _from_pyarrow(
        cls: Type[_E],
        array: pa.Array,
        schema: pt.PortexType,
        root: Optional["DataFrame"] = None,
        name: Tuple[str, ...] = (),
    ) -> _E:
        obj: _E = object.__new__(cls)
        obj._data = PyArrowPagingList.from_pyarrow(array)
        obj.schema = schema
        obj._root = root
        obj._name = name
        obj._indices_to_values = cls._create_indices_to_values(
            schema.to_builtin().values,  # type: ignore[attr-defined]
            obj._data._patype,  # pylint: disable=protected-access
        )
        return obj

    @staticmethod
    def _create_indices_to_values(
        enum_values: List[Union[int, float, str, bool, None]], patype: pa.DataType
    ) -> Dict[Optional[pa.Scalar], Union[int, float, str, bool, None]]:
        indices_to_values = dict(
            zip(
                pa.array(range(len(enum_values)), patype),  # pylint: disable=protected-access
                enum_values,
            )
        )
        if None not in indices_to_values:
            indices_to_values[pa.scalar(None, patype)] = None
        return indices_to_values

    def to_pylist(self) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        return [self._indices_to_values[i] for i in self._data]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Functions about uploading files."""

from copy import deepcopy
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterable, Tuple

import filetype
from requests_toolbelt import MultipartEncoder
from tqdm import tqdm

from graviti.openapi.data import get_policy
from graviti.openapi.requests import do
from graviti.utility import File, config, convert_iso_to_datetime, locked, submit_multithread_tasks

_EXPIRED_IN_SECOND = 240

_PERMISSIONS: Dict[Tuple[str, str, int], Dict[str, Any]] = {}
_DEFAULT_PERMISSION = {"expire_at": datetime.fromtimestamp(0, timezone.utc)}


@locked
def _request_upload_permission(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
) -> None:
    permission = get_policy(
        access_key,
        url,
        owner,
        dataset,
        draft_number=draft_number,
        sheet=sheet,
        is_internal=config.is_internal,
        expired=_EXPIRED_IN_SECOND,
    )
    permission["expire_at"] = convert_iso_to_datetime(permission.pop("expire_at"))

    _PERMISSIONS[access_key, dataset, draft_number] = permission


def _get_upload_permission(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
) -> Dict[str, Any]:
    key = (access_key, dataset, draft_number)
    if datetime.now(timezone.utc) >= _PERMISSIONS.get(key, _DEFAULT_PERMISSION)["expire_at"]:
        _request_upload_permission(
            access_key,
            url,
            owner,
            dataset,
            draft_number=draft_number,
            sheet=sheet,
        )

    return deepcopy(_PERMISSIONS[key])


def upload_files(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    files: Iterable[File],
    jobs: int,
    pbar: tqdm,
) -> None:
    """Upload multiple local files with multithread.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        files: The local files to upload.
        jobs: The number of the max workers in multi-thread uploading procession.
        pbar: The process bar for uploading binary files.

    """
    submit_multithread_tasks(
        lambda file: upload_file(
            access_key=access_key,
            url=url,
            owner=owner,
            dataset=dataset,
            draft_number=draft_number,
            sheet=sheet,
            file=file,
            pbar=pbar,
        ),
        files,
        jobs=jobs,
    )


def upload_file(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    file: File,
    pbar: tqdm,
) -> None:
    """Upload one local file.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        file: The local file to upload.
        pbar: The process bar for uploading binary files.

    """
    permission = _get_upload_permission(
        access_key,
        url,
        owner,
        dataset,
        draft_number=draft_number,
        sheet=sheet,
    )

    post_data = permission["result"]

    local_path = file.path
    checksum = file.checksum

    permission_extra = permission["extra"]
    post_data["key"] = permission_extra["object_prefix"] + checksum

    host = permission_extra["host"]
    backend_type = permission_extra["backend_type"]

    if backend_type == "azure":
        url = (
            f'{permission_extra["host"]}{permission_extra["object_prefix"]}'
            f'{checksum}?{permission["result"]["token"]}'
        )
        _put_binary_file_to_azure(url, local_path, post_data)

    elif backend_type == "fps":
        _post_multipart_formdata(
            host,
            local_path,
            post_data,
            checksum,
        )
    elif backend_type == "s3":
        if "signedRequestHeaders" in post_data:
            del post_data["signedRequestHeaders"]

        _post_multipart_formdata(
            host,
            local_path,
            post_data,
        )
    else:
        _post_multipart_formdata(
            host,
            local_path,
            post_data,
        )

    pbar.update()


def _post_multipart_formdata(
    url: str, local_path: Path, data: Dict[str, Any], file_name: str = ""
) -> None:
    with local_path.open("rb") as fp:
        file_type = filetype.guess_mime(local_path)
        if "x-amz-date" in data:
            data["Content-Type"] = file_type

        data["file"] = (file_name, fp, file_type)
        multipart = MultipartEncoder(data)
        do(
            "POST",
            url,
            data=multipart,
            headers={"Content-Type": multipart.content_type},
        )


def _put_binary_file_to_azure(url: str, local_path: Path, data: Dict[str, Any]) -> None:
    with local_path.open("rb") as fp:
        file_type = filetype.guess_mime(local_path)
        request_headers = {
            "x-ms-blob-content-type": file_type,
            "x-ms-blob-type": data["x-ms-blob-type"],
        }
        do("PUT", url, data=fp, headers=request_headers)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""OpenAPI module."""

from graviti.openapi.branch import create_branch, delete_branch, get_branch, list_branches
from graviti.openapi.commit import commit_draft, get_commit, get_revision, list_commits
from graviti.openapi.data import add_data, list_commit_data, list_draft_data, update_data
from graviti.openapi.dataset import (
    create_dataset,
    delete_dataset,
    get_dataset,
    list_datasets,
    update_dataset,
)
from graviti.openapi.draft import create_draft, get_draft, list_drafts, update_draft
from graviti.openapi.file import upload_files
from graviti.openapi.schema import update_schema
from graviti.openapi.search import create_search
from graviti.openapi.sheet import (
    create_sheet,
    delete_sheet,
    get_commit_sheet,
    get_draft_sheet,
    list_commit_sheets,
    list_draft_sheets,
)
from graviti.openapi.tag import create_tag, delete_tag, get_tag, list_tags
from graviti.openapi.user import get_current_user

__all__ = [
    "add_data",
    "commit_draft",
    "create_branch",
    "create_dataset",
    "create_draft",
    "create_search",
    "create_sheet",
    "create_tag",
    "delete_branch",
    "delete_dataset",
    "delete_sheet",
    "delete_tag",
    "get_branch",
    "get_commit",
    "get_commit_sheet",
    "get_current_user",
    "get_dataset",
    "get_draft",
    "get_draft_sheet",
    "get_revision",
    "get_tag",
    "list_branches",
    "list_commit_data",
    "list_commit_sheets",
    "list_commits",
    "list_datasets",
    "list_draft_data",
    "list_draft_sheets",
    "list_drafts",
    "list_tags",
    "update_data",
    "update_dataset",
    "update_draft",
    "update_schema",
    "upload_files",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Dataframe module."""

from graviti.dataframe.column.series import FileSeries
from graviti.dataframe.column.series import SeriesBase as ColumnSeriesBase
from graviti.dataframe.container import Container
from graviti.dataframe.frame import RECORD_KEY, DataFrame

__all__ = ["Container", "DataFrame", "FileSeries", "RECORD_KEY", "ColumnSeriesBase"]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the dataset policy."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def get_object_policy(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    actions: Optional[str] = None,
    is_internal: Optional[bool] = None,
    expired: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/policy/object`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        actions: The specific actions including "GET" and "PUT". The default value in the OpenAPI
            is "GET". Supports multiple actions, which need to be separated by ``|``, like
            "GET|PUT".
        is_internal: Whether to return the intranet upload address, the default value in
            the OpenAPI is False.
        expired: Token expiry time in seconds. It cannot be negative.

    Returns:
        The response of OpenAPI.

    Examples:
        Request permission to get dataset data:

        >>> get_policy(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ... )
        {
            "backend_type":"OSS",
            "policy": {
                "AccessKeyId":"LTAI4FjgXD3yFJUaasdasd",
                "AccessKeySecret":"LTAI4FjgXD3yFJJKasdad",
                "SecurityToken":"CAISrgJ1q6Ft5B2yfSjIr5bkKILdaseqw",
                "bucket":"content-store-dev",
                "endpoint":"content-store-dev.oss-cn-qingdao.aliyuncs.com",
                "expireAt":"2022-07-12T06:07:52Z"
            }
        }

        Request permission to put dataset data:

        >>> get_policy(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     actions="PUT",
        ... )
        {
            "backend_type":"OSS",
            "policy": {
                "AccessKeyId":"LTAI4FjgXD3yFJUaasdasd",
                "AccessKeySecret":"LTAI4FjgXD3yFJJKasdad",
                "SecurityToken":"CAISrgJ1q6Ft5B2yfSjIr5bkKILdaseqw",
                "bucket":"content-store-dev",
                "endpoint":"content-store-dev.oss-cn-qingdao.aliyuncs.com",
                "expireAt":"2022-07-12T06:07:52Z",
                "prefix":"051dd0676cc74f548a7e9b7ace45c26b/"
            }
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/policy/object"
    params: Dict[str, Any] = {}

    if actions is not None:
        params["actions"] = actions
    if is_internal is not None:
        params["is_internal"] = is_internal
    if expired is not None:
        params["expired"] = expired

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]
